using System;
using System.Drawing;
using System.Windows.Forms;

using System.IO;

// single C# file WinForm application
// save as MnistDistort.cs
// to compile: >csc.exe /target:winexe MistDistort.cs

// try kernel dim = 13 k stddev = 3.0 intensity = 18.0
// image data can be found at http://yann.lecun.com/exdb/mnist/

namespace MnistDistort
{

  static class Program
  {
    [STAThread]
    static void Main()
    {
      Application.EnableVisualStyles();
      Application.SetCompatibleTextRenderingDefault(false);
      Application.Run(new Form1());
    }
  }

  public partial class Form1 : Form
  {
    DigitImage[] trainImages = null;
    string pixelFile = @"C:\MnistDistort\train-images.idx3-ubyte"; // EDIT THIS AS NEEDED!
    string labelFile = @"C:\MnistDistort\train-labels.idx1-ubyte"; // EDIT THIS AS NEEDED!

    public Form1()
    {
      InitializeComponent();
      textBox1.Text = pixelFile;
      textBox2.Text = labelFile;
      comboBox1.SelectedItem = "6";  // Magnification
      textBox3.Text = "NA"; // curr index
      textBox4.Text = "0"; // next index
      this.ActiveControl = button1;
    }

    // UI code ======================================

    private void button1_Click(object sender, EventArgs e) // load MNIST train data
    {
      string pixelFile = textBox1.Text;
      string labelFile = textBox2.Text;
      trainImages = LoadData(pixelFile, labelFile);
      listBox1.Items.Add("MNIST training images loaded into memory");
    }

    private void button2_Click(object sender, EventArgs e)
    {
      // display 'next' image
      int nextIndex = int.Parse(textBox4.Text);
      DigitImage currImage = trainImages[nextIndex];

      int mag = int.Parse(comboBox1.SelectedItem.ToString()); // magnification
      Bitmap bitMap = MakeBitmap(currImage, mag);
      pictureBox1.Image = bitMap;

      textBox3.Text = textBox4.Text; // update curr index from old next index
      textBox4.Text = (nextIndex + 1).ToString(); // ++next index
      textBox8.Text = textBox3.Text;  // use the curr indexs as a seed for Displacement

      listBox1.Items.Add("Curr image index = " + textBox3.Text + " label = " + currImage.label);
    }

    private void button3_Click(object sender, EventArgs e) // distort curr image
    {
      int currIndex = int.Parse(textBox3.Text);
      int mag = int.Parse(comboBox1.SelectedItem.ToString());
      int kDim = int.Parse(textBox5.Text); // Gaussian kernel dimension
      double kStdDev = double.Parse(textBox6.Text); // kernel std dev
      double intensity = double.Parse(textBox7.Text);
      int rndSeed = int.Parse(textBox8.Text);  // use the curr image index (or manual) as seed

      DigitImage currImage = trainImages[currIndex];
      DigitImage distorted = Distort(currImage, kDim, kStdDev, intensity, rndSeed);
      Bitmap bitMapDist = MakeBitmap(distorted, mag);
      pictureBox2.Image = bitMapDist;
    }

    // Logic code ===================================

    private static DigitImage[] LoadData(string pixelFile, string labelFile)
    {
      // Load MNIST training set of 60,000 images into memory
      int numImages = 60000; // ugly
      DigitImage[] result = new DigitImage[numImages];

      byte[][] pixels = new byte[28][];
      for (int i = 0; i < pixels.Length; ++i)
        pixels[i] = new byte[28];

      FileStream ifsPixels = new FileStream(pixelFile, FileMode.Open);
      FileStream ifsLabels = new FileStream(labelFile, FileMode.Open);

      BinaryReader brImages = new BinaryReader(ifsPixels);
      BinaryReader brLabels = new BinaryReader(ifsLabels);

      int magic1 = brImages.ReadInt32(); // read and discard
      int imageCount = brImages.ReadInt32();
      int numRows = brImages.ReadInt32();
      int numCols = brImages.ReadInt32();

      int magic2 = brLabels.ReadInt32();
      int numLabels = brLabels.ReadInt32();

      // each image
      for (int di = 0; di < numImages; ++di)
      {
        for (int i = 0; i < 28; ++i) // get 28x28 pixel values
        {
          for (int j = 0; j < 28; ++j)
          {
            byte b = brImages.ReadByte();
            pixels[i][j] = b;
          }
        }

        byte lbl = brLabels.ReadByte(); // get the label
        DigitImage dImage = new DigitImage(28, 28, pixels, lbl);
        result[di] = dImage;
      } // each image

      ifsPixels.Close();
      brImages.Close();
      ifsLabels.Close();
      brLabels.Close();

      return result;
    } // LoadData

    private static Bitmap MakeBitmap(DigitImage dImage, int mag)
    {
      // create a C# Bitmap suitable for display in a PictureBox control
      int width = dImage.width * mag;
      int height = dImage.height * mag;
      Bitmap result = new Bitmap(width, height);
      Graphics gr = Graphics.FromImage(result);
      for (int i = 0; i < dImage.height; ++i)
      {
        for (int j = 0; j < dImage.width; ++j)
        {
          int pixelColor = 255 - dImage.pixels[i][j]; // white background, black digits
          //int pixelColor = dImage.pixels[i][j]; // black background, white digits
          Color c = Color.FromArgb(pixelColor, pixelColor, pixelColor);
          SolidBrush sb = new SolidBrush(c);
          gr.FillRectangle(sb, j * mag, i * mag, mag, mag); // fills bitmap via Graphics object
        }
      }
      return result;
    }

    private DigitImage Distort(DigitImage dImage, int kDim, double kStdDev, double intensity, int seed)
    {
      double[][] kernel = MakeKernel(kDim, kStdDev);
      double[][] xField = MakeDisplace(dImage.width, dImage.height, seed, kernel, intensity);
      double[][] yField = MakeDisplace(dImage.width, dImage.height, seed + 1, kernel, intensity);
      byte[][] newPixels = Displace(dImage.pixels, xField, yField);
      return new DigitImage(dImage.width, dImage.height, newPixels, dImage.label);
    }

    private static double[][] MakeKernel(int dim, double sd)
    {
      // make a dim x dim continuous Gaussian kernel
      if (dim % 2 == 0) throw new Exception("kernel dim must be odd");

      double[][] result = new double[dim][];
      for (int i = 0; i < dim; ++i)
        result[i] = new double[dim];

      int center = dim / 2; // note truncation
      double coef = 1.0 / (2.0 * Math.PI * (sd * sd)); // normalizes so all value sum to (close) 1.0
      double denom = 2.0 * (sd * sd);
      double sum = 0.0; // for more accurate normalization
      for (int i = 0; i < dim; ++i)
      {
        for (int j = 0; j < dim; ++j)
        {
          int x = Math.Abs(center - j);
          int y = Math.Abs(center - i);
          double num = -1.0 * ((x * x) + (y * y));
          double z = coef * Math.Exp(num / denom);
          result[i][j] = z;
          sum += z;
        }
      }

      for (int i = 0; i < dim; ++i)
        for (int j = 0; j < dim; ++j)
          result[i][j] = result[i][j] / sum;

      return result;
    } // MakeKernel 

    private static double[][] MakeDisplace(int width, int height, int seed, double[][] kernel, double intensity)
    {
      // first, generate a (-1, +1) displacement [in either the x- or y-direction]
      // then apply a Gaussian kernel
      // then multiply by an intensity factor
      double[][] dField = new double[height][];
      for (int i = 0; i < dField.Length; ++i)
        dField[i] = new double[width];

      Random rnd = new Random(seed); // want different displacements! 

      for (int i = 0; i < dField.Length; ++i)
        for (int j = 0; j < dField[i].Length; ++j)
          dField[i][j] = 2.0 * rnd.NextDouble() - 1.0; // (1.0 - -1.0) * rnd + -1.0

      dField = ApplyKernel(dField, kernel); // smooth displacement vectors. .
 
      for (int i = 0; i < dField.Length; ++i)
        for (int j = 0; j < dField[i].Length; ++j)
          dField[i][j] *= intensity;

      return dField;
    }

    private static double[][] ApplyKernel(double[][] dField, double[][] kernel)
    {
      // 'convolve' displacement field with a kernel
      int borderAdd = kernel.Length / 2; // ex: for 7x7 need 3 extra rows top bottom left right
      double[][] padded = Pad(dField, borderAdd, borderAdd, borderAdd, borderAdd);

      double[][] result = new double[dField.Length][];
      for (int i = 0; i < result.Length; ++i)
        result[i] = new double[dField[i].Length];

      for (int i = 0; i < dField.Length; ++i) // i pts to dField row
      {
        for (int j = 0; j < dField[i].Length; ++j) // j = dField col
        {
          double newValue = 0.0; // new value based on old val and 8 neighbors

          for (int kr = 0, pr = i; kr < kernel.Length; ++kr, ++pr) // kr = kernel row
          {
            for (int kc = 0, pc = j; kc < kernel.Length; ++kc, ++pc)
            {
              newValue += kernel[kr][kc] * padded[pr][pc]; // pr = padded row (dField vals)
            }
          }
          result[i][j] = newValue;
        }
      }
      return result;
    } // ApplyKernel

    //static double[][] ApplyKernel(double[][] matrix, double[][] kernel, bool noPadding)
    //{
    //  // matrix values 'out-of-range' are left alone
    //  int n = kernel.Length / 2; // ex: for 7x7 need 3 extra rows top bottom left right. assumes square kernel

    //  double[][] result = new double[matrix.Length][];
    //  for (int i = 0; i < result.Length; ++i)
    //    result[i] = new double[matrix[i].Length];

    //  for (int i = 0; i < matrix.Length; ++i) // i pts to matrix row
    //  {
    //    for (int j = 0; j < matrix[i].Length; ++j) // j = matrix col
    //    {
    //      if (i - n < 0 || i + n >= matrix.Length || j - n < 0 || j + n >= matrix[i].Length)
    //      {
    //        result[i][j] = matrix[i][j];
    //        continue; // bad i or j
    //      }

    //      double newValue = 0.0;

    //      for (int kr = 0, mr = i - n; kr < kernel.Length; ++kr, ++mr) // kernel row, matrix row
    //      {
    //        for (int kc = 0, mc = j - n; kc < kernel.Length; ++kc, ++mc) // kernel col
    //        {
    //          double prod = (kernel[kr][kc] * matrix[mr][mc]);
    //          newValue += prod;
    //        } // kr
    //      } // kc
    //      //result[i][j] = (int)Math.Round(newValue, 0);
    //      result[i][j] = newValue;

    //    } // j
    //  } // i
    //  return result;
    //}

    private static double[][] Pad(double[][] dField, int top, int bottom, int left, int right)
    {
      int oRows = dField.Length; // original number rows
      int oCols = dField[0].Length;
      int nRows = oRows + top + bottom; // new (padded) number rows
      int nCols = oCols + left + right;

      double[][] result = new double[nRows][];
      for (int i = 0; i < nRows; ++i)
        result[i] = new double[nCols];

      for (int i = top; i < oRows + top; ++i)
      {
        for (int j = left; j < oCols + left; ++j)
        {
          result[i][j] = dField[i - top][j - left]; // copy orig values to center area
        }
      }
      return result;
    } // Pad

    private static byte[][] Displace(byte[][] pixels, double[][] xField, double[][] yField)
    {
      // take each pixel value and replace it with a val detemined by the x- and y- fields

      byte[][] resultPixels = new byte[pixels.Length][];
      for (int i = 0; i < pixels.Length; ++i)
        resultPixels[i] = new byte[pixels[i].Length];

      //int picker = 0;
      for (int i = 0; i < pixels.Length; ++i)
      {
        for (int j = 0; j < pixels[i].Length; ++j)
        {
          //int ii = i + (int)yField[i][j];
          //int jj = j + (int)xField[i][j];

          int lowii = i + (int)Math.Floor(yField[i][j]);
          int hiii = i + (int)Math.Ceiling(yField[i][j]);
          int lowjj = j + (int)Math.Floor(xField[i][j]);
          int hijj = j + (int)Math.Ceiling(xField[i][j]);

          if (lowii < 0 || lowjj < 0 || hiii > pixels.Length - 1 || hijj > pixels[i].Length - 1)
            resultPixels[i][j] = 0;
          else
          {
            int sum = pixels[lowii][lowjj] + pixels[lowii][hijj] + pixels[hiii][lowjj] + pixels[hiii][hijj];
            double avg = Math.Round(sum / 4.0);
            byte avgVal = (byte)avg;
            resultPixels[i][j] = avgVal; // avg value produces a smoothing effect
          }
        }
      }
      return resultPixels;
    }

    // ------------------ VS-generated code

    private System.Windows.Forms.TextBox textBox1;
    private System.Windows.Forms.TextBox textBox2;
    private System.Windows.Forms.PictureBox pictureBox1;
    private System.Windows.Forms.PictureBox pictureBox2;
    private System.Windows.Forms.ListBox listBox1;
    private System.Windows.Forms.Button button1;
    private System.Windows.Forms.TextBox textBox3;
    private System.Windows.Forms.TextBox textBox4;
    private System.Windows.Forms.ComboBox comboBox1;
    private System.Windows.Forms.Button button2;
    private System.Windows.Forms.Label label1;
    private System.Windows.Forms.Label label2;
    private System.Windows.Forms.Label label3;
    private System.Windows.Forms.TextBox textBox5;
    private System.Windows.Forms.TextBox textBox6;
    private System.Windows.Forms.TextBox textBox7;
    private System.Windows.Forms.Label label4;
    private System.Windows.Forms.Label label5;
    private System.Windows.Forms.Label label6;
    private System.Windows.Forms.Label label7;
    private System.Windows.Forms.Button button3;
    private System.Windows.Forms.TextBox textBox8;
    private System.Windows.Forms.Label label8;

    private System.ComponentModel.IContainer components = null;
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }
     private void InitializeComponent()
    {
      this.textBox1 = new System.Windows.Forms.TextBox();
      this.textBox2 = new System.Windows.Forms.TextBox();
      this.pictureBox1 = new System.Windows.Forms.PictureBox();
      this.pictureBox2 = new System.Windows.Forms.PictureBox();
      this.listBox1 = new System.Windows.Forms.ListBox();
      this.button1 = new System.Windows.Forms.Button();
      this.textBox3 = new System.Windows.Forms.TextBox();
      this.textBox4 = new System.Windows.Forms.TextBox();
      this.comboBox1 = new System.Windows.Forms.ComboBox();
      this.button2 = new System.Windows.Forms.Button();
      this.label1 = new System.Windows.Forms.Label();
      this.label2 = new System.Windows.Forms.Label();
      this.label3 = new System.Windows.Forms.Label();
      this.textBox5 = new System.Windows.Forms.TextBox();
      this.textBox6 = new System.Windows.Forms.TextBox();
      this.textBox7 = new System.Windows.Forms.TextBox();
      this.label4 = new System.Windows.Forms.Label();
      this.label5 = new System.Windows.Forms.Label();
      this.label6 = new System.Windows.Forms.Label();
      this.label7 = new System.Windows.Forms.Label();
      this.button3 = new System.Windows.Forms.Button();
      this.textBox8 = new System.Windows.Forms.TextBox();
      this.label8 = new System.Windows.Forms.Label();
      ((System.ComponentModel.ISupportInitialize)(this.pictureBox1)).BeginInit();
      ((System.ComponentModel.ISupportInitialize)(this.pictureBox2)).BeginInit();
      this.SuspendLayout();
      // 
      // textBox1
      // 
      this.textBox1.Location = new System.Drawing.Point(12, 22);
      this.textBox1.Name = "textBox1";
      this.textBox1.Size = new System.Drawing.Size(308, 20);
      this.textBox1.TabIndex = 0;
      // 
      // textBox2
      // 
      this.textBox2.Location = new System.Drawing.Point(12, 48);
      this.textBox2.Name = "textBox2";
      this.textBox2.Size = new System.Drawing.Size(308, 20);
      this.textBox2.TabIndex = 1;
      // 
      // pictureBox1
      // 
      this.pictureBox1.BackColor = System.Drawing.SystemColors.ControlDark;
      this.pictureBox1.Location = new System.Drawing.Point(12, 143);
      this.pictureBox1.Name = "pictureBox1";
      this.pictureBox1.Size = new System.Drawing.Size(280, 280);
      this.pictureBox1.TabIndex = 2;
      this.pictureBox1.TabStop = false;
      // 
      // pictureBox2
      // 
      this.pictureBox2.BackColor = System.Drawing.SystemColors.Desktop;
      this.pictureBox2.Location = new System.Drawing.Point(326, 143);
      this.pictureBox2.Name = "pictureBox2";
      this.pictureBox2.Size = new System.Drawing.Size(280, 280);
      this.pictureBox2.TabIndex = 3;
      this.pictureBox2.TabStop = false;
      // 
      // listBox1
      // 
      this.listBox1.FormattingEnabled = true;
      this.listBox1.Location = new System.Drawing.Point(12, 443);
      this.listBox1.Name = "listBox1";
      this.listBox1.Size = new System.Drawing.Size(280, 108);
      this.listBox1.TabIndex = 4;
      // 
      // button1
      // 
      this.button1.Location = new System.Drawing.Point(326, 32);
      this.button1.Name = "button1";
      this.button1.Size = new System.Drawing.Size(95, 23);
      this.button1.TabIndex = 5;
      this.button1.Text = "Load Images";
      this.button1.UseVisualStyleBackColor = true;
      this.button1.Click += new System.EventHandler(this.button1_Click);
      // 
      // textBox3
      // 
      this.textBox3.Enabled = false;
      this.textBox3.Location = new System.Drawing.Point(12, 91);
      this.textBox3.Name = "textBox3";
      this.textBox3.Size = new System.Drawing.Size(62, 20);
      this.textBox3.TabIndex = 6;
      // 
      // textBox4
      // 
      this.textBox4.Location = new System.Drawing.Point(80, 91);
      this.textBox4.Name = "textBox4";
      this.textBox4.Size = new System.Drawing.Size(62, 20);
      this.textBox4.TabIndex = 7;
      // 
      // comboBox1
      // 
      this.comboBox1.FormattingEnabled = true;
      this.comboBox1.Items.AddRange(new object[] {
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10"});
      this.comboBox1.Location = new System.Drawing.Point(159, 91);
      this.comboBox1.Name = "comboBox1";
      this.comboBox1.Size = new System.Drawing.Size(62, 21);
      this.comboBox1.TabIndex = 8;
      // 
      // button2
      // 
      this.button2.Location = new System.Drawing.Point(245, 91);
      this.button2.Name = "button2";
      this.button2.Size = new System.Drawing.Size(75, 23);
      this.button2.TabIndex = 9;
      this.button2.Text = "Disply Next";
      this.button2.UseVisualStyleBackColor = true;
      this.button2.Click += new System.EventHandler(this.button2_Click);
      // 
      // label1
      // 
      this.label1.AutoSize = true;
      this.label1.Location = new System.Drawing.Point(9, 114);
      this.label1.Name = "label1";
      this.label1.Size = new System.Drawing.Size(55, 13);
      this.label1.TabIndex = 10;
      this.label1.Text = "Curr Index";
      // 
      // label2
      // 
      this.label2.AutoSize = true;
      this.label2.Location = new System.Drawing.Point(77, 114);
      this.label2.Name = "label2";
      this.label2.Size = new System.Drawing.Size(58, 13);
      this.label2.TabIndex = 11;
      this.label2.Text = "Next Index";
      // 
      // label3
      // 
      this.label3.AutoSize = true;
      this.label3.Location = new System.Drawing.Point(156, 115);
      this.label3.Name = "label3";
      this.label3.Size = new System.Drawing.Size(70, 13);
      this.label3.TabIndex = 12;
      this.label3.Text = "Magnification";
      // 
      // textBox5
      // 
      this.textBox5.Location = new System.Drawing.Point(553, 55);
      this.textBox5.Name = "textBox5";
      this.textBox5.Size = new System.Drawing.Size(53, 20);
      this.textBox5.TabIndex = 13;
      this.textBox5.Text = "13";
      // 
      // textBox6
      // 
      this.textBox6.Location = new System.Drawing.Point(553, 81);
      this.textBox6.Name = "textBox6";
      this.textBox6.Size = new System.Drawing.Size(53, 20);
      this.textBox6.TabIndex = 14;
      this.textBox6.Text = "3.0";
      // 
      // textBox7
      // 
      this.textBox7.Location = new System.Drawing.Point(553, 107);
      this.textBox7.Name = "textBox7";
      this.textBox7.Size = new System.Drawing.Size(53, 20);
      this.textBox7.TabIndex = 15;
      this.textBox7.Text = "18.0";
      // 
      // label4
      // 
      this.label4.AutoSize = true;
      this.label4.Location = new System.Drawing.Point(555, 9);
      this.label4.Name = "label4";
      this.label4.Size = new System.Drawing.Size(51, 13);
      this.label4.TabIndex = 16;
      this.label4.Text = "Distortion";
      // 
      // label5
      // 
      this.label5.AutoSize = true;
      this.label5.Location = new System.Drawing.Point(487, 58);
      this.label5.Name = "label5";
      this.label5.Size = new System.Drawing.Size(60, 13);
      this.label5.TabIndex = 17;
      this.label5.Text = "Kernel Size";
      // 
      // label6
      // 
      this.label6.AutoSize = true;
      this.label6.Location = new System.Drawing.Point(466, 84);
      this.label6.Name = "label6";
      this.label6.Size = new System.Drawing.Size(79, 13);
      this.label6.TabIndex = 18;
      this.label6.Text = "Kernel Std Dev";
      // 
      // label7
      // 
      this.label7.AutoSize = true;
      this.label7.Location = new System.Drawing.Point(466, 110);
      this.label7.Name = "label7";
      this.label7.Size = new System.Drawing.Size(81, 13);
      this.label7.TabIndex = 19;
      this.label7.Text = "Alpha (intensity)";
      // 
      // button3
      // 
      this.button3.Location = new System.Drawing.Point(346, 91);
      this.button3.Name = "button3";
      this.button3.Size = new System.Drawing.Size(75, 23);
      this.button3.TabIndex = 20;
      this.button3.Text = "Distort Curr";
      this.button3.UseVisualStyleBackColor = true;
      this.button3.Click += new System.EventHandler(this.button3_Click);
      // 
      // textBox8
      // 
      this.textBox8.Location = new System.Drawing.Point(553, 29);
      this.textBox8.Name = "textBox8";
      this.textBox8.Size = new System.Drawing.Size(53, 20);
      this.textBox8.TabIndex = 21;
      this.textBox8.Text = "0";
      // 
      // label8
      // 
      this.label8.AutoSize = true;
      this.label8.Location = new System.Drawing.Point(471, 32);
      this.label8.Name = "label8";
      this.label8.Size = new System.Drawing.Size(76, 13);
      this.label8.TabIndex = 22;
      this.label8.Text = "Displace Seed";
      // 
      // Form1
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(632, 570);
      this.Controls.Add(this.label8);
      this.Controls.Add(this.textBox8);
      this.Controls.Add(this.button3);
      this.Controls.Add(this.label7);
      this.Controls.Add(this.label6);
      this.Controls.Add(this.label5);
      this.Controls.Add(this.label4);
      this.Controls.Add(this.textBox7);
      this.Controls.Add(this.textBox6);
      this.Controls.Add(this.textBox5);
      this.Controls.Add(this.label3);
      this.Controls.Add(this.label2);
      this.Controls.Add(this.label1);
      this.Controls.Add(this.button2);
      this.Controls.Add(this.comboBox1);
      this.Controls.Add(this.textBox4);
      this.Controls.Add(this.textBox3);
      this.Controls.Add(this.button1);
      this.Controls.Add(this.listBox1);
      this.Controls.Add(this.pictureBox2);
      this.Controls.Add(this.pictureBox1);
      this.Controls.Add(this.textBox2);
      this.Controls.Add(this.textBox1);
      this.Name = "Form1";
      this.Text = "MNIST Distorter";
      ((System.ComponentModel.ISupportInitialize)(this.pictureBox1)).EndInit();
      ((System.ComponentModel.ISupportInitialize)(this.pictureBox2)).EndInit();
      this.ResumeLayout(false);
      this.PerformLayout();
    } // InitializeComponent()

    

    // ------------------



  } // class Form1

  public class DigitImage
  {
    // an MNIST image of a '0' thru '9' digit
    public int width; // 28
    public int height; // 28
    public byte[][] pixels; // 0(white) - 255(black)
    public byte label; // '0' - '9'

    public DigitImage(int width, int height, byte[][] pixels, byte label)
    {
      this.width = width; this.height = height;
      this.pixels = new byte[height][];
      for (int i = 0; i < this.pixels.Length; ++i)
        this.pixels[i] = new byte[width];

      for (int i = 0; i < height; ++i)
        for (int j = 0; j < width; ++j)
          this.pixels[i][j] = pixels[i][j];

      this.label = label;
    }
  }

  

  

} // ns

